/* local sorting and bitonic merge routines
 */

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/bulk.h>
#include <split-c/com.h>
#include "modes.h"
#include "macros.h"
#include "sort.h"

extern int data[PROCS]::[MAX_PER_PROC];
static check[PROCS]::;

printData(localData, k, message)
doubleword *localData;
int	k;
char *message;
{
  int ii;
  int jj;
  /* print initial */
  on_one {
    for(jj = 0; jj<PROCS; jj++){
      printf(" > > > PROCESSOR %i < < <\n", jj);
      for(ii = 0; ii < k; ii++){
	printf("\t%010i\n", data[jj][ii]);
      }	
    }
  }
}

void checkData(int k)
{
  int ii,jj;
  on_one{
    printf("checking correctness of sort\n");
  }
  for(ii = 0; ii < (k-1); ii++){
    if (data[MYPROC][ii] < data[MYPROC][ii+1])
      printf("error at proc %d item %d\n",MYPROC,ii);
  }
  check[MYPROC]=data[MYPROC][k-1];
  barrier();
  if (MYPROC < (PROCS - 1)) {
    if (check[MYPROC] < check[MYPROC+1]) {
      printf("proc %d: error across to  %d \n",MYPROC,MYPROC+1);  
    }
  }
}

/* returns who processor iproc should exchange data with when the inner 
   loop is working on dimension innderdim */
int getBuddy(iproc, innerdim)
int iproc;
int innerdim;
{
    if(iproc & (BITMASK(innerdim))){
	return(iproc - (OFFSET(innerdim)));
    }
    else{
	return(iproc + (OFFSET(innerdim)));
    }
}


int getMode(iproc, innerdim, outerdim)
int iproc;
int innerdim;
int outerdim;
{
    /* xor of bit (innerdim) of iproc and bit outerdim of iproc */
    int innerbit, outerbit;

    innerbit = iproc & (BITMASK(innerdim));
    outerbit = iproc & (BITMASK(outerdim));

    if((innerbit || outerbit) && !(innerbit && outerbit)){
	return COMP_MIN;
    }
    else{
	return COMP_MAX;
    }

}


int sortMode(iproc, outerdim)
int iproc;
int outerdim;
{
    int outerbit;

    /* ASCENDING if bit outerdim of iproc is 1
       DESCENDING if bit outerdim of iproc is 0 */
    outerbit = iproc & (BITMASK(outerdim));

    if(outerbit){
	return ASCENDING;
    }
    else{
	return DESCENDING;
    }
}


localSort(data, n, mode)
doubleword *data;
int n;
int mode;
{
    if(mode == ASCENDING){
	upSort(data, n);
    }
    else{
	downSort(data, n);
    }
}

/* quicksort version */

/* partion around randomly chosen element. */
/* left half <= P, right half >= P.  Progress guaranteed b/c
   we don't recurse on P */
/* not stable, but we don't care in this algorithm.  */
int
quickUpPartition( data, n)
doubleword *data;
int n;
{
  int pp;
  doubleword p;
  register int front, back;
  doubleword temp;

  pp = rand() % n;

  assign(p, data[pp]);
  
  /* don't move p --we need to move it into partion position */
  front = 0; 
  back = n-1;
  while(front < back){
    while(is_lessorequal_than(data[front], p) && front < back){
      front++;
    }
    while(is_greaterorequal_than(data[back], p) && back > front){
      back--;
    }
    if(front < back){ /* swap */
      assign(temp, data[back]);
      assign(data[back], data[front]);
      assign(data[front], temp);
    }
  }
  /* front and back now point to same element--we don't know if it
     is < = > than p */

  if(is_lessorequal_than(data[front], p)){ 
    front++;
  }
  else{
    back--;
  }


  /* front == back+1 
     everything from 0..back is <= p
     everything from front..n-1 is >= p
     Need to be careful of weird case where back == -1 front == 0 
                                      or    back == n-1 front = n */
  /* if the partition element is to left of where we met, we need to
     swap it and the element under back. If it to the right we need
     to swap it with the element under front.  If it is under back
     of fornt, we need to return it. */
  if(pp < back){  /* swap data[back], data[pp]  */
    assign(temp, data[back]);
    assign(data[back], p);
    assign(data[pp], temp);
    return back;
  }
  else if (pp > front){
    assign(temp, data[front]);
    assign(data[front], p);
    assign(data[pp], temp);
    return front;
  }
  else if (pp == front){
    return front;
  }
  else if (pp == back){
    return back;
  }
  else{
    printf("Bug in partion code.  Cant find pp\n");
    exit(1);
  }
}


int
quickDownPartition( data, n)
doubleword *data;
int n;
{
  int pp;
  doubleword p;
  register int front, back;
  doubleword temp;

  pp = rand() % n;
  assign(p, data[pp]);

  /* don't move p --we need to move it into partion position */
  front = 0; 
  back = n-1;
  while(front < back){
    while(is_greaterorequal_than(data[front], p) && front < back){
      front++;
    }
    while(is_lessorequal_than(data[back], p) && back > front){
      back--;
    }
    if(front < back){ /* swap */
      assign(temp, data[back]);
      assign(data[back], data[front]);
      assign(data[front], temp);
    }
  }
  /* front and back now point to same element--we don't know if it
     is < = > than p */

  if(is_greaterorequal_than(data[front], p)){ 
    front++;
  }
  else{
    back--;
  }

  /* front == back+1 
     everything from 0..back is >= p
     everything from front..n-1 is <= p
     Need to be careful of weird case where back == -1 front == 0 
                                      or    back == n-1 front = n */
  /* if the partition element is to left of where we met, we need to
     swap it and the element under back. If it to the right we need
     to swap it with the element under front.  If it is under back
     of fornt, we need to return it. */

  if(pp < back){  /* swap data[back], data[pp]  */
    assign(temp, data[back]);
    assign(data[back], p);
    assign(data[pp], temp);
    return back;
  }
  else if (pp > front){
    assign(temp, data[front]);
    assign(data[front], p);
    assign(data[pp], temp);
    return front;
  }
  else if (pp == front){
    return front;
  }
  else if (pp == back){
    return back;
  }
  else{
    printf("Bug in partion code.  Cant find pp\n");
    exit(1);
  }
}

upSort(data, n)
doubleword *data;
int n;
{
  int p;

  if( n <= 1 ) {
    return;
  }
  p = quickUpPartition( data, n);  /* p is index of partition element */
  upSort(data, p);  /* sort from begining up to but not incl. p */
  upSort( data[p + 1], n-p-1); /* sort from after p to end */

}



downSort(data, n)
doubleword *data;
int n;
{
  int p;

  if( n <= 1 ) {
    return;
  }
  p = quickDownPartition( data, n);  /* p is index of partition element */
  downSort(data, p);
  downSort( data[p + 1], n-p-1);

}





