/*									tab:4
 *
 * column.sc - Implementation of column sort algorithm.  
 *
 * "Copyright (c) 1993 and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea C. Dusseau
 * Version:			143
 * Creation Date:	Tue Jun 29 10:53:10 1993
 * Filename:		column.sc
 * History:
 *	LTL	142	Sat Dec 17 13:22:08 1994
 *		Changed bcopy to memcpy and random to rand
 */
static char _version_string_[] = "\nVersion:142:column.sc\0\n";

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include "column.h"

#define R_MAX (1024*1024)
int matrix1[PROCS]::[R_MAX];
int matrix2[PROCS]::[R_MAX];
int *col1, *col2;

int R = 8 * 1024;
int N;

int TransAlg = 0;
int UntrAlg = 0;

int tr_offset;
int BAND;

#define RADIX          65536 
#define RADIX_MIN_1    65535
#define LOG_RADIX         16



/* a 8 bit bucket and a 32 bit bucket if the first bucket overflows */
static unsigned char  buckets[RADIX];
static unsigned int   buck[RADIX];

void inline
radix_sort_up(int *c, int *t)
{
    unsigned int *x;
    int i, j, sum, temp;

    /* Init both buckets */
    for (j = 0; j < RADIX; ++j) {
	buckets[j] = 0;
	buck[j] = 0;
    }

    /* if 1st bucket overflows icr the 2nd bucket */
    for (j = 0; j < R; j++) {
	if ((++buckets[*(((unsigned short *)&c[j])+1)]) == 0)
	    buck[*(((unsigned short *)&c[j])+1)]++;
    }
    
    sum = 0;
    for (j = 0; j < RADIX; ++j) {
	temp = buckets[j] + buck[j]*256;
	buck[j] = sum;
	sum += temp;
    }
    
    for (j = 0; j < R; j++) {
	sum = buck[*(((unsigned short *)&c[j])+1)]++;
	t[sum] = c[j];
    }

    x = t;
    t = c;
    c = x;

    /* Init both buckets */
    for (j = 0; j < RADIX; ++j) {
	buckets[j] = 0;
	buck[j] = 0;
    }

    for (j = 0; j < R; j++) {
	if ( (++buckets[*((unsigned short *)&c[j])]) == 0 )
	    buck[*((unsigned short *)&c[j])]++;
    }
    
    sum = 0;
    for (j = 0; j < RADIX; ++j) {
	temp = buckets[j] + buck[j]*256;
	buck[j] = sum;
	sum += temp;
    }

    for (j = 0; j < R; j++) {
	sum = buck[*((unsigned short *)&c[j])]++;
	t[sum] = c[j];
    }
}

void merge(int *srcBuf, int *destBuf)
{
  int m, end1, end2;
  int n, k;

  m = 0;
  end1 = n = m + (R>>1);
  end2 = end1 + (R>>1);

  /* loop for the merge - you break out as soon as you reach the 
     end of one of the two buffers being merged - no check in for loop */
  for (k = 0; ; ) {

    if (srcBuf[m] <= srcBuf[n]) {
      destBuf[k++] = srcBuf[m++];
      if ( m == end1 ) break;
    }
    else {
      destBuf[k++] = srcBuf[n++];
      if ( n == end2 ) break;
    }
  }

  /* Check whether m got over first or n got over first */
  if ( m == end1 )
    for ( ; n < end2 ; )
      destBuf[k++] = srcBuf[n++];
  else
    for ( ; m < end1 ; )
      destBuf[k++] = srcBuf[m++];
}

void InitCols(int *col)
{
  int i;
  
  for (i = 0; i < R; i++)
    col[i] = rand(); 
}

double t[10];

void splitc_main(int argc, char *argv[])
{
  int c;
  extern char *optarg;

  /* 
   * Get options from the user 
   */
  while ((c = getopt(argc, argv, "t:u:r:")) != -1) {
    switch(c) {
    case 't':  TransAlg = atoi(optarg); break;
    case 'u':  UntrAlg = atoi(optarg); break;
    case 'r':  R = atoi(optarg);  R *= 1024; break;
    }
  }

  /*
   * Check options -- number of keys is constrained.
   */
  if (R % PROCS) {
    on_one {
      printf("ERROR: R (%d) must be evenly divisible by the number of processors.\n", R);
      exit(-1);
    }
  }
  if (R < 2 * (PROCS-1) * (PROCS-1)) {
    on_one {
      printf("ERROR: R (%d) must be >= 2 * (PROCS-1)^2 (%d).\n", R, 
	     2 * (PROCS-1) * (PROCS-1));
      exit(-1);
    }
  }
  if (R > R_MAX) 
    on_one {
      printf("ERROR: R (%d) must be <= %d\n", R, R_MAX);
      exit(-1);
    }
  N = R * PROCS;

  /*
   * Print out configuration 
   */
  on_one {
    printf("\nSorting %d keys on %d processors (%d keys/proc).\n", N, PROCS, R);
    printf("Transpose algorithm: ");
    switch (TransAlg) {
    case 0:
      printf("Store keys individually.\n");
      break;
    case 1:
      printf("Store keys individually -- spacing proc destinations.\n");
      break;
    case 2:
      printf("Gather keys and store contiguosly.\n");
      break;
    case 3:
      printf("Gather keys and store contiguosly -- spacing proc destinations.\n");
      break;
    default:
      printf("Unknown algorithm.\n");
      exit(-1);
    }
    printf("Untranspose algorithm: ");
    switch (UntrAlg) {
    case 0:
      printf("Store keys contiguously.\n");
      break;
    case 1:
      printf("Store keys contiguously -- spacing proc destinations.\n");
      break;
    default:
      printf("Unknown algorithm.\n");
      exit(-1);
    }
#ifdef ALL_STORE
    printf("Using all_store_sync()\n");
#else
    printf("Using store_sync() with expected number of bytes.\n");
#endif

#ifdef SYNC_EACH
    printf("Synchronizing after each store operation.\n");
#else
    printf("Synchronizing after each transpose operation.\n");
#endif
  
  }


  col1 = (int *)matrix1;
  col2 = (int *)matrix2;
  /*
   * Do the column sort algorithm, timing individual components. . .
   */
      
  InitCols(col1);
  barrier();

  t[0] = get_seconds();

  t[2] = get_seconds();
  radix_sort_up(col1, col2);
  t[2] = get_seconds()-t[2];

  barrier(); 
  t[3] = get_seconds();
  BAND = R / PROCS;
  tr_offset = MYPROC * BAND;
  Transpose(col2, col1);
  t[3] = get_seconds() - t[3];

  t[4] = get_seconds();
  radix_sort_up(col2, col1);
  t[4] = get_seconds() - t[4];

  t[5] = get_seconds();
  barrier();
  Untranspose(col1, col2);
  t[5] = get_seconds() - t[5];


  t[6] = get_seconds();
  radix_sort_up(col1, col2);
  t[6] = get_seconds() - t[6];

  t[7] = get_seconds();
  barrier();
  Shift(col2, col1);
  t[7] = get_seconds() - t[7];

  t[8] = get_seconds();
  if (MYPROC) merge(col2,col1);
  else memcpy(col1, col2, R*sizeof(int));
  t[8] = get_seconds() - t[8];

  t[9] = get_seconds();
  barrier();
  Unshift(col2, col1);
  t[9] = get_seconds() - t[9];

  t[0] = get_seconds()-t[0];

  PrintResults();
  CheckResults(col2, 1);
}


void PrintResults()
{
  int i;
  double max, min, sum;

#if 0
  for (i = 0; i < 10; i++) {
    if (i == 1) continue;

    max = all_reduce_to_one_dmax(t[i]);
    min = all_reduce_to_one_dmin(t[i]);
    sum = all_reduce_to_one_dadd(t[i]);
    on_one  {
      switch (i) {
      case 0: printf("\nTotal --"); break;
      case 2: printf("1.  Sort --"); break;
      case 3: printf("2.  Trans --"); break;
      case 4: printf("3.  Sort --"); break;
      case 5: printf("4.  Untrans --"); break;
      case 6: printf("5.  Sort --"); break;
      case 7: printf("6.  Shift --"); break;
      case 8: printf("7.  Merge --"); break;
      case 9: printf("8.  Unshift --"); break;
      }
      printf("\tMin: %.3lf s \tMax: %.3lf s \tAve: %.3lf s\n", min, max, sum/PROCS);
    }
  } 
#else
  on_one printf("Procs: %d   Keys/PROCS: %d   ", PROCS, R);
  for (i = 0; i < 10; i++) {
    if (i == 1) continue;
    
    sum = all_reduce_to_one_dadd(t[i]);
    on_one  {
      switch (i) {
      case 0: printf("Total --"); break;
      case 2: printf("Sort --"); break;
      case 3: printf("Trans --"); break;
      case 4: printf("Sort --"); break;
      case 5: printf("Untrans --"); break;
      case 6: printf("Sort --"); break;
      case 7: printf("Shift --"); break;
      case 8: printf("Merge --"); break;
      case 9: printf("Unshift --"); break;
      }
      printf(" %.3lf s   ", sum/PROCS);
    }
  } 
  on_one printf("\n");
#endif

}


void Shift(int *d, int *s)
{
  int *global dst;

  memcpy(d, s, sizeof(int)*(R>>1));

  dst = toglobal(MYPROC != PROCS-1 ? MYPROC+1 : 0, d + (R>>1));
  bulk_store(dst, s + (R>>1), sizeof(int)*(R>>1));

#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*(R>>1));
#endif
    
}

void Unshift(int *d, int *src)
{
  int *global dst;

  if (!MYPROC) {
    memcpy(d, src, sizeof(int)*(R>>1));
    dst = toglobal(PROCS-1, d + (R>>1));
    bulk_store(dst, src + (R>>1), sizeof(int)*(R>>1));

  }
  else {
    memcpy(d, src + (R>>1), sizeof(int)*(R>>1));
    dst = toglobal(MYPROC-1, d + (R>>1));
    bulk_store(dst, src, sizeof(int)*(R>>1));
  }

#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*(R>>1));
#endif

}

void Transpose(int *d, int *s)
{
  switch (TransAlg) {
  case 0:
    TransSingle(d, s); break;
  case 1:
    TransSingleOpt(d, s); break;
  case 2:
    TransBuf(d, s); break;
  case 3:
    TransBufOpt(d, s); break;
  }
}

/*
 * Naively store each local element into its correct destination.  Everyone is
 * slamming on the same processors in the same order. . .
 */
void TransSingle(int *d, int *s)
{
  int *global dest;
  int i, p, *index;

  index = d + tr_offset;
  for (i = 0, p = 0; i < R; i++, p = (p+1)%PROCS) {
    dest = toglobal(p, index + i/PROCS);
    *dest :- *s++;
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int));
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}

void TransSingleOpt(int *d, int *s)
{
  int *global dest;
  int i, p, j, *index;

  index = d + tr_offset;
  for (p = MYPROC, i = 0, j = MYPROC; i < R; i++, p = (p+1)%PROCS, j = (j+1)%R) {
    dest = toglobal(p, index + j/PROCS);
    *dest :- *s++;
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int));
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}

/*
 * Buffer each scattered local element destined for the same processor into a temporary 
 * local array.  Then store this contiguous array on the destination processor.
 */ 
void TransBuf(int *d, int *s)
{
  int *global dst;
  int temp[PROCS][BAND];
  int p, j, i, *index;

  index = d + tr_offset;
  for (p = 0; p < PROCS; p++) {
    for (i = 0, j = p; j < R; i++, j += PROCS)
      temp[p][i] = s[j];
    dst = toglobal(p, index);
    bulk_store(dst, temp[p], sizeof(int) * BAND);
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int)*BAND);
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}


void TransBufOpt(int *d, int *s)
{
  int *global dst;
  int temp[PROCS][BAND];
  int p, p2, j, i, *index;

  index = d + tr_offset;
  for (p2 = 0, p = MYPROC; p2 < PROCS; p2++, p = (p+1)%PROCS) {
    for (i = 0, j = p; j < R; i++, j += PROCS)
      temp[p][i] = s[j];
    dst = toglobal(p, index);
    bulk_store(dst, temp[p], sizeof(int) * BAND);
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int)*BAND);
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}

void Untranspose(int *d, int *s)
{
  switch (UntrAlg) {
  case 0:
    Untrans(d, s); break;
  case 1:
    UntransOpt(d, s); break;
  }
}

void Untrans(int *d, int *s)
{
  int *global dst;
  int p, i, *index;

  index = d + tr_offset;
  for (i = 0, p = 0; p < PROCS; p++, i+=BAND) {
    dst = toglobal(p, index);
    bulk_store(dst, s + i, sizeof(int) * BAND);
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int)*BAND);
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}



void UntransOpt(int *d, int *s)
{
  int *global dst;
  int p, i, j, *index;

  index = d + tr_offset;
  for (j = 0, p = MYPROC; j < PROCS; j++, p = (p+1)%PROCS) {
    dst = toglobal(p, index);
    bulk_store(dst, s + (p*BAND), sizeof(int) * BAND);
#ifdef SYNC_EACH
#ifdef ALL_STORE
    all_store_sync();
#else
    store_sync(sizeof(int)*BAND);
#endif
#endif
  }
#ifndef SYNC_EACH
#ifdef ALL_STORE
  all_store_sync();
#else
  store_sync(sizeof(int)*R);
#endif
#endif
}

static check[PROCS]::;

void CheckResults(int *b, int flag)
{
  int ii,jj;
  int error = 0;

  on_one printf("Checking correctness of sort. . .\n");

  for(ii = 0; ii < (R-1); ii++){
    if (b[ii] > b[ii+1]) {
      printf("Error at proc %d item %d\n",MYPROC,ii);
      error++;
    }
  }

  if (flag) {
  check[MYPROC] = b[R-1];
  barrier();
  if (MYPROC < (PROCS - 1)) {
    if (check[MYPROC] > check[MYPROC+1]) {
      printf("proc %d: error across to  %d \n",MYPROC,MYPROC+1);  
      error++;
    }
  }
}
  if (!error) on_one printf(". . .keys sorted correctly.\n");
}


  
      

  
















