/*									tab:8
 *
 *
 * 
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:				Arvind
 * Version:				54
 * Creation Date:		
 * Filename:			
 * History:
 */

#include <stdio.h>
#include <split-c/split-c.h>
#include "connect.h"

/*
   What to do:
   1. find the biggest component -- local.sc
   2. do connected components on the biggest component
   3. label that component the lowest numbered component
*/

/*
   Input is the maximum-sized component on each processor.
   Overall strategy:  do S-V on PROCS vertices.  Dont move
   edges.  Emulate CRCW architecture -- the processors have
   nothing else to do.  Another restriction -- the values are
   in the range 0...PROCS-1.
*/

extern double max_conn_time;

void
all_max_conn_comp(node_t *max_comp)
{
    int done, complete, inorder;
    edge_t *global *e_ptr, *edge;

    max_conn_time = get_seconds();
    inorder = 1;
    /* Preprocessing to get the relevant edges in front */
    e_ptr = &max_comp->edges;
    edge = tolocal(*e_ptr);
    while (edge) {
	int other_val = edge->node->value;
	if (other_val < PROCS) {
	    if (!inorder) {
		/* move to beginning of list */
		*e_ptr = edge->next;
		edge->next = max_comp->edges;
		max_comp->edges = (edge_t *global)edge;
		edge = tolocal(*e_ptr);
	    } else {
		/* skip to next element */
		e_ptr = &edge->next;
		edge = tolocal(edge->next);
	    }
	} else {
	    /* found the first not-relevant edge */
	    inorder = 0;
	    e_ptr = &edge->next;
	    edge = tolocal(edge->next);
	}
    }
    barrier();

    done = 0;
    while (1) {
	int my_val = max_comp->value;
	edge_t *alt_edge = NULL;
	    
	/* Hooking phase... */
	e_ptr = &max_comp->edges;
	edge = tolocal(*e_ptr);
	while (edge) {
	    int other_val = edge->node->value;

	    /* reached the end of relevant edges */
	    if (other_val >= PROCS) {
		edge = NULL;
		break;
	    }

	    /* a duplicate edge */
	    if (other_val == my_val) {
		/* delete the edge */
		*e_ptr = edge->next;
		edge = tolocal(edge->next);
	    }
	    else {
		/* a hookable edge */
		if (other_val < my_val) break;

		/* a non-hookable edge is remembered for later use */
		if (!alt_edge) {
		    alt_edge = edge;
		}
		
		e_ptr = &edge->next;
		edge = tolocal(edge->next);
	    }
	}

	if (edge != NULL) {
	    node_t *global parent = max_comp->parent;
	    node_t *global other_node = edge->node;

	    /* This is the CW part of the algorithm. */
	    parent->parent = other_node;
	}
	barrier();

#if 0
	if (edge == NULL && alt_edge) {
	    /* Unconditional hooking... */
	    node_t *global parent = max_comp->parent;
	    
	    /* Okay, parent didnt hook */
	    if (parent->parent == parent) {
		node_t *global other_node = alt_edge->node;

		/* update parent link */
		parent->parent = other_node;
	    }
	}
#endif	
	barrier();

	/* Maximal Pointer doubling... */
	for (;;) {
	    node_t *global parent = max_comp->parent;
	    node_t *global gp = parent->parent;
	    int completed;

	    if (parent->parent == parent) {
		max_comp->value = parent->value;
		completed = all_reduce_to_all_add(1);
	    }
	    else {
		max_comp->parent = gp;
		completed = all_reduce_to_all_add(0);
	    }
	    if (completed == PROCS) break;
	} 

	done |= ((max_comp->edges == NULL) ||
		 (max_comp->edges->node->value >= PROCS));
	
	/* Termination condition... */
	complete = all_reduce_to_all_add (done);
	if (complete == PROCS) break;
    }
#if 0    
    {
	int num_left = all_reduce_to_all_add(max_comp->parent == (node_t *global)max_comp);
	if (num_left == 1) {
	    max_comp->edges = (edge_t *global)NULL;
	}
	else if (num_left < PROCS) {
	    printf("not sure what to do...\n");
	}
    }
#endif
    max_conn_time = get_seconds() - max_conn_time;
}
