/*									tab:4
 *
 * em3d.h - Header file for the electromagnetic problem in three dimensions.
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Steve Lumetta
 * Version:			11
 * Creation Date:	Sat Jan 23 12:22:55 1993
 * Filename:		em3d.h
 * History:
 *	SL	1	Sat Jan 23 12:42:25 1993
 *		Prepared for release as example (without proper balancing).
 */

#ifndef _EM3D_H
#define _EM3D_H

#include <split-c/split-c.h>

#undef EXTERN
#undef INIT
#ifdef _EM3D_MAIN
#define EXTERN
#define INIT(a) = (a)
#else
#define EXTERN extern
#define INIT(a)
#endif

/* Cost function parameters */
#define COST_LOCAL 0
#define COST_SEND 2000
#define COST_REMOTE 200
#define COST_CALC 25
#define HIGH_COST 2147483647

/* Time step parameters */
#define TIME_STEPS 100
#define DELTA_T 0.1

/* Data type for nodes */
typedef double DATA;

/* List of nodes dependent on a node */
typedef struct to_node_list_t {
    int length;
    struct node_t *global *nodes;
} to_node_list_t;

/* List of nodes on which a node depends (and coefficients) */
typedef struct from_node_list_t {
    int length;
    DATA* coeffs;
	union {
        struct node_t *global *nodes;
#ifdef EM_SIMPLE     /* Simple version uses global data pointers */
		DATA *global *values;
#else
		DATA **values;
#endif		
    } nv;
} from_node_list_t;

/* Node structure */
typedef struct node_t {
    int locked;
    DATA value;
    struct to_node_list_t to_nodes;
    struct from_node_list_t from_nodes;
    struct node_t **prev;
    struct node_t *next;
} node_t;

typedef struct ghost_node_t {
	DATA value;
	DATA *global rval;
	struct ghost_node_t *next;
} ghost_node_t;
	
/* Prototypes of functions used to obtain original graph */
void all_initialize_graph();

/* Prototypes of functions used to move nodes for load balancing */
int check_percent(int chance);
int gen_number(int limit);
int get_cost(int proc_num);
void grab_h_node(node_t *global a_node);
void grab_e_node(node_t *global a_node);

/* Prototypes of functions used to calculate solution */
void all_prepare_packing();
void all_compute_e();
void all_compute_h();
int splitc_main();

/* Global variables */
EXTERN int n_h_nodes;
EXTERN int n_e_nodes;
EXTERN node_t *h_nodes;
EXTERN node_t *e_nodes;
EXTERN node_t *dump_list INIT(NULL);

EXTERN ghost_node_t *h_ghost_nodes;
EXTERN ghost_node_t *e_ghost_nodes;
	
EXTERN int h_pack_len[PROCSMAX];
EXTERN int g_h_rcv_len[PROCS]::[PROCSMAX];
EXTERN int *h_rcv_len;
EXTERN int h_rcv_total;
EXTERN DATA **h_pack_list[PROCSMAX];
EXTERN DATA *h_pack_area[PROCSMAX];
EXTERN DATA *global g_h_put_area[PROCS]::[PROCSMAX];
EXTERN DATA *global *h_put_area;
EXTERN DATA *h_rcv_area[PROCSMAX];

EXTERN int e_pack_len[PROCSMAX];
EXTERN int g_e_rcv_len[PROCS]::[PROCSMAX];
EXTERN int *e_rcv_len;
EXTERN int e_rcv_total;
EXTERN DATA **e_pack_list[PROCSMAX];
EXTERN DATA *e_pack_area[PROCSMAX];
EXTERN DATA *global g_e_put_area[PROCS]::[PROCSMAX];
EXTERN DATA *global *e_put_area;
EXTERN DATA *e_rcv_area[PROCSMAX];

EXTERN int g_cost[PROCS]::[PROCSMAX];
EXTERN int *cost;
EXTERN int *my_cost;

#undef EXTERN
#undef INIT

#endif

