
/*
 *  FFT -- Split-C FFT program
 */  

#include <stdio.h>
#include <split-c/split-c.h>
#include <math.h>
#include <split-c/com.h>
#include "defs.h"

#define	BARRIER		/* if defined, put barriers every few stores */

int terse = 1;
int i, j;

foo(int n, complex IN[n]::, complex OUT[n]::) {
	IN[i] = OUT[j];
}

int check(int k,int n,int c_size,complex OUT[PROCS]::[n/PROCS])
{ 
  int trouble=0,i; 
  for (i=0; i<c_size; i++ ) {
    if ( ((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i)->re!= 2.0 )
      trouble++;
  }
  if (trouble) printf("check %d: troubles=%d\n",k,trouble);
}

int printin(int n, complex IN[n]::)
{
  int i,j;
  complex *start = tolocal(IN);

  for (i=0; i< PROCS; i++) {
    barrier();
    if (MYPROC==i) {
      for (j=0; j<n/PROCS; j++) {
	printf("IN[%d]=%.2f,%.2f (addr=%d)\n",j,
	       (start+j)->re,(start+j)->im,start+j);
      }
    }
  }
}

int printout(int n, complex OUT[PROCS]::[n/PROCS])
{
  int i,j;
  complex *start = tolocal(OUT);

  for (i=0; i< PROCS; i++) {
    barrier();
    if (MYPROC==i) {
      for (j=0; j<n/PROCS; j++) {
	printf("OUT[%d]=%.2f,%.2f (addr=%d)\n",j, 
	       (start+j)->re,(start+j)->im,start+j);
      }
    }
  }
}
 
int setin(int n, complex IN[n]::)
{
  int i,j;
  complex *start = tolocal(IN);

  for (i=0; i< PROCS; i++) {
    barrier();
    if (MYPROC==i) {
      for (j=0; j<n/PROCS; j++) {
	(start+j)->re= MYPROC;
	(start+j)->im= j;
      }
    }
  }
}

int setout(int n, complex OUT[PROCS]::[n/PROCS])
{
  int i,j;
  complex *start = tolocal(OUT);

  for (i=0; i< PROCS; i++) {
    barrier();
    if (MYPROC==i) {
      for (j=0; j<n/PROCS; j++) {
	(start+j)->re= (double)MYPROC/100;
	(start+j)->im= (double)j/100;
      }
    }
  }
}
 
void
all_remap_cyclic_to_blocked(int n, complex IN[n]::,
							complex OUT[PROCS]::[n/PROCS])
{
	int i, j, blk_size, c_size;
	complex *local, *aux, *top;
	double ttime[4], tt;
	double t,t2;
	
	barrier();
	blk_size = n/PROCS;
	c_size = blk_size/PROCS;
	local = tolocal(&(IN[blk_size*MYPROC + MYPROC]));
	top = tolocal(&(IN[MYPROC]));	

	printf("PROCS=%d n=%d blk_size=%d c_size=%d\n",
	       PROCS,n,blk_size,c_size);
	printf("local in at addr %d.\n",local);
	printf("start setting %d numbers\n",c_size);
	for (i=0; i<c_size; i++ ) {
	  (local+i)->re= MYPROC+1;
	  (local+i)->im= i+1;
	}
	setin(n,IN);
	setout(n,OUT);
	barrier();
	printin(n,IN);
	printout(n,OUT);
	barrier();

#if 1
	t=get_seconds();

	for (i=0; i<c_size; i++ ) {
	  ((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i)->re= 2.0;
	  ((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i)->im= 0.0;
	}
	t2=get_seconds();
	printf("touch t=%f (bytes=%d) @ addr=%d\n",t2-t,
	       c_size*sizeof(complex),
	       (complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) ); 
#endif
	tt = get_seconds();
	
	/* stagger the communication */
	/* send the last part first */
	for (i=MYPROC; i < PROCS; i++)
	{
		t=get_seconds();
printf("from:%d to:(%d,%d) size:%d\n",
       local + (i-MYPROC)*c_size,
       toproc(OUT[i] + c_size*i),
       tolocal(OUT[i] + c_size*i),
       c_size*sizeof(complex));
		bulk_store(OUT[i] + c_size*i, local + (i-MYPROC)*c_size, c_size*sizeof(complex));
		/* sync();		 */
		barrier();
		t2=get_seconds();
		printf("t=%f\n",t2-t);
	}
/*	all_store_sync(); */
	check(2,n,c_size,OUT);

	/* now send the rest, starting from the top */
	for (i=0; i < MYPROC; i++)
	{
		t=get_seconds();
printf("from:%d to:(%d,%d) size:%d\n",
       local + (i-MYPROC)*c_size,
       toproc(OUT[i] + c_size*i),
       tolocal(OUT[i] + c_size*i),
       c_size*sizeof(complex));
		bulk_store(OUT[i] + c_size*i, top + i*c_size, c_size*sizeof(complex));
		/* sync();		*/
		barrier(); 
		t2=get_seconds();
		printf("t=%f\n",t2-t);
	}
	barrier();

	tt = get_seconds() - tt;
	ttime[0] = get_seconds();
	

	printout(n,OUT);
	local = tolocal(OUT[MYPROC]);
	aux = tolocal(&(IN[MYPROC]));
	
	check(3,n,c_size,OUT);

	for (i=0; i < blk_size; i++)
		aux[i/c_size + (i%c_size)*PROCS] = local[i];

	memcpy(local, aux, blk_size*sizeof(complex));	

	ttime[0] = get_seconds() - ttime[0];
	on_one printf("Remap inside  in %.6fs (%.3fMb/s/proc)\n", tt,
			(blk_size*sizeof(complex))/(1E6*tt));
	on_one printf("Transfer time in %.6fs (%.3fMb/s/proc)\n",ttime[0], (blk_size*sizeof(complex))/(1E6*ttime[0]));


}

#if 0 
void
all_remap_cyclic_to_blocked(int n, complex IN[n]::,
			    complex OUT[PROCS]::[n/PROCS])
{
	int i, j, blk_size, c_size;
	complex *local, *aux, *top;
	double ttime[4], tt;
	double t,t2;
	
	barrier();
	blk_size = n/PROCS;
	c_size = blk_size/PROCS;
	local = tolocal(&(IN[blk_size*MYPROC + MYPROC]));
	top = tolocal(&(IN[MYPROC]));	

	printf("PROCS=%d n=%d blk_size=%d c_size=%d\n",
	       PROCS,n,blk_size,c_size);
	printf("local in at addr %d.\n",local);
	printf("start setting %d numbers\n",c_size);
	for (i=0; i<c_size; i++ ) {
	  printf("setting complex IN number at address %d \n",(local+i));
	  (local+i)->re= MYPROC+1;
	  (local+i)->im= i+1;
	}
	barrier();
	printin(n,IN);

	t=get_seconds();

	for (i=0; i<c_size; i++ ) {
	  printf("setting complex OUT number at address %d \n",
		((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i) );
	  ((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i)->re= 2.0;
	  ((complex *)tolocal(OUT[MYPROC] + c_size*MYPROC) +i)->im= 0.0;
	}
	check(1,n,c_size,OUT);
	t2=get_seconds();
	printf("touch t=%f (bytes=%d\n",t2-t,c_size*sizeof(complex)); 


	{
	int i, begn, blk_sz;
	complex *local;

	barrier();
	blk_sz = n/PROCS;
	begn = (blk_sz*MYPROC/PROCS)*PROCS + MYPROC;
	local = tolocal(&(IN[begn]));
	for (i=begn; i<n; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	local = tolocal(&IN[MYPROC]);
	for (i=MYPROC; i!=begn; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	barrier();
      }
	printout(n,OUT);

}
#endif

#if 0
/*--------------------------------------------------------- original */
void
all_remap_cyclic_to_blocked(int n, complex IN[n]::,
			    complex OUT[PROCS]::[n/PROCS])
{
	int i, begn, blk_sz;
	complex *local;

	barrier();
	blk_sz = n/PROCS;
	begn = (blk_sz*MYPROC/PROCS)*PROCS + MYPROC;
	local = tolocal(&(IN[begn]));
	for (i=begn; i<n; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	local = tolocal(&IN[MYPROC]);
	for (i=MYPROC; i!=begn; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	barrier();
}

/*--------------------------------------------------------- */
#endif

double ttime;
static void
start_timer(void)
{
	ttime = get_seconds();
	barrier();
}

static double
stop_timer(void)
{
	barrier();
	return get_seconds() - ttime;
}

void
all_fft(unsigned int n, complex X[n]::)
{
	int i;
	unsigned int my_n = n/PROCS;
	complex *my_X = (complex *)X;
	complex (*spread T)[n/PROCS];
	complex TF[my_n/2];
	double t;
	extern int flop;

	T = tospread(0, (complex *)all_spread_malloc(n, sizeof(complex)));

	/* X starts with cyclic mapping (i.e. as declared)
	 * Do log(n)-log(PROCS) local FFT steps */
	if(!terse && MYPROC == 0) printf("Starting %d-point local FFT\n", my_n);
	start_timer(); flop = 0;
	local_fft(my_n, my_n/2, (complex *)X, TF);
	t = stop_timer();
	if(MYPROC == 0)
		printf("Phase 1 in %.6fs (%.3fMflops/proc)\n", t, flop/(1E6*t));

	/* Permute X from cyclic mapping to block mapping */
	if(!terse && MYPROC == 0) printf("Starting remapping\n");
	start_timer();
	all_remap_cyclic_to_blocked(n, X, (void *spread)T);
	t = stop_timer();
	if(MYPROC == 0)
		printf("Remap   in %.6fs (%.3fMb/s/proc)\n", t,
			(my_n*sizeof(complex))/(1E6*t));

	/* T is now blocked mapping.
	 * Do log(PROCS) local FFT steps */
	if(!terse && MYPROC == 0)
		printf("Starting %d %d-point local FFTs\n", my_n/PROCS, PROCS);
	start_timer(); flop = 0;
	local_fft(my_n, PROCS/2, (complex *)T, TF);
	t = stop_timer();
	if(MYPROC == 0)
		printf("Phase 2 in %.6fs (%.3fMflops/proc)\n", t, flop/(1E6*t));
}

void splitc_main(int argc, char * argv[])
{
	int n = 1, i;
	complex *spread X;
	char c;	
	extern char *optarg;
	
	/* determine and broadcast FFT size */

	/* parse command line args */
	while ((c = getopt(argc, argv, "n:h")) != -1) {
		switch(c) {
		case 'h':
			on_one{
				printf("FFT options: (defaults)\n");
				printf("  -n mil. of keys (in K) per processor  (1 mil)\n");
			}
			exit(0);

		case 'n':  n = atoi(optarg); break;
		}
	}

/*	n = (1 << 10) * n;	*/
		
	on_one
	{	
		printf("Doing %d point FFT on %d processors \n", n, PROCS);
		printf("On each processor I have %d points and I allocate %d Kbytes\n", n/PROCS, (n/PROCS*32)>>10);		
	}
	
	/* all_store_sync(); */
	/* n = all_bcast_i(n); */

	/* allocate and init input array */
	X = (complex *spread)all_spread_malloc(n, sizeof(complex));
	barrier();
	if (!X) {
		if (MYPROC == 0) fprintf(stderr, "Not enough memory!\n");
		barrier();
		exit(1);
	}
	
	for(i=0; i<n/PROCS; i++)
		((complex *)X)[i] = (complex){0.0, 0.0};
	barrier();
	if(!terse && MYPROC == 0) printf("Initialization done\n");

	/* do FFT */
	all_fft(n, X);
	barrier();

	if(!terse && MYPROC == 0) printf("Done.\n");
}

