/*									tab:4
 * fft.sc - Split-C FFT with a combination of cyclic and blocked layout
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Thorsten von Eicken
 * Version:			64
 * Creation Date:	Fri Oct  2 22:21:44 PDT 1992
 * Filename:		fft.sc
 * History:
 *	AK	9	Wed Mar 16 17:32:58 1994
 *		Wrote a new version of all_remap to study performance improvements
 */

#include <stdio.h>
#include <split-c/split-c.h>
#include <math.h>
/* #include <cm/cmmd.h> chris */
#include "defs.h"

#if 0
#define	BARRIER		/* if defined, put barriers every few stores */
#endif

int terse = 1;

int i, j;


#if 0
void
all_remap_cyclic_to_blocked(int n, complex_t IN[n]::,
							complex_t OUT[PROCS]::[n/PROCS])
{
	int i, begn, blk_sz;
	complex_t *local;

	barrier();
	blk_sz = n/PROCS;
	begn = (blk_sz*MYPROC/PROCS)*PROCS + MYPROC;
	local = tolocal(&(IN[begn]));
	for (i=begn; i<n; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	local = tolocal(&IN[MYPROC]);
	for (i=MYPROC; i!=begn; i+=PROCS)
		OUT[i/blk_sz][i%blk_sz] :- *local++;

	barrier();
}

void
all_remap_cyclic_to_blocked(int n, complex_t IN[n]::,
							complex_t OUT[PROCS]::[n/PROCS])
{
	int i, begn, o, p;
	int blk_sz = n/PROCS;
	complex_t *local;
	complex_t *global remote;

	barrier();
	/* Take each element owned by me and store it on the
	   destination processor that requires the element.  */

	/* Start from my processor */
	begn = (blk_sz*MYPROC/PROCS)*PROCS + MYPROC;
	local = tolocal(IN + begn);
	o = begn % blk_sz;
	p = begn / blk_sz;
	remote = OUT[p];
	for (; p<PROCS; ) {
		remote[o] :- *(local++);
		o += PROCS;
		while (o >= blk_sz) {
			o -= blk_sz;
			p++;
			remote = OUT[p];
			sync();
			barrier();
		}
	}

	/* Wrap around */
	local = tolocal(IN + MYPROC);
	o = MYPROC % blk_sz;
	p = MYPROC / blk_sz;
	remote = OUT[p];
	for (i=MYPROC; i!=begn; i += PROCS) {
		remote[o] :- *(local++);
		o += PROCS;
		while (o >= blk_sz) {
			o -= blk_sz;
			p++;
			remote = OUT[p];
			sync();
			barrier();
		}
	}
	all_store_sync();
}
#endif

#if 1
void
all_remap_cyclic_to_blocked(int n,
	complex_t IN[n]::, complex_t OUT[PROCS]::[n/PROCS])
{
	int i, o, p;
	complex_t *my_IN = (complex_t *)IN;
	complex_t *in;
	complex_t *out = (complex_t *)OUT[0];
	unsigned int block_sz = n/PROCS;
	unsigned int start;

	/* start by storing onto self */
	start = block_sz*MYPROC/PROCS;

	barrier();
	o = (start*PROCS+MYPROC) % block_sz;
	p = (start*PROCS+MYPROC) / block_sz;
#if 1
	{
		complex_t *global my_OUT = &OUT[p][0];
		for(i=start, in=my_IN+start; i<block_sz; i++,in++) {
			my_OUT[o] :- my_IN[i];
			o += PROCS;
			while(o >= block_sz) {
#ifdef BARRIER
				barrier();
#endif
				p++; o -= block_sz;
				my_OUT = OUT[p];
			}
		}
	}
#else
	for(i=start, in=my_IN+start; i<block_sz; i++,in++) {
		OUT[p][o] = my_IN[i];
		o += PROCS;
		while(o >= block_sz) {
#ifdef BARRIER
			barrier();
#endif
			p++; o -= block_sz;
		}
	}
#endif
	o = MYPROC % block_sz;
	p = MYPROC / block_sz;
	for(i=0,in=my_IN; i<start; i++,in++) {
		OUT[p][o] :- my_IN[i];
		o += PROCS;
		while(o >= block_sz) {
#ifdef BARRIER
			barrier();
#endif
			p++; o -= block_sz;
		}
	}

	barrier();
/*	all_store_sync(); */
}
#endif

double ttime;
static void
start_timer(void)
{
/*	
	if(MYPROC == 0) {
		CMMD_node_timer_clear(0);
		barrier();
		CMMD_node_timer_start(0);
	} else {
		barrier();
	}
*/
	ttime = get_seconds();
	barrier();
}

static double
stop_timer(void)
{
/*	
	if(MYPROC == 0) {
		barrier();
		CMMD_node_timer_stop(0);
		return CMMD_node_timer_busy(0);
	} else {
		barrier();
		return 0.0;
	}
*/
	barrier();
	return get_seconds() - ttime;
}

void
all_fft(unsigned int n, complex_t X[n]::)
{
	int i;
	unsigned int my_n = n/PROCS;
	complex_t *my_X = (complex_t *)X;
	complex_t (*spread T)[n/PROCS];
	complex_t TF[my_n/2];
	double t;
	extern int flop;

	T = tospread(0, (complex_t *)all_spread_malloc(n, sizeof(complex_t)));

	/* X starts with cyclic mapping (i.e. as declared)
	 * Do log(n)-log(PROCS) local FFT steps */
	if(!terse && MYPROC == 0) printf("Starting %d-point local FFT\n", my_n);
	barrier();
	start_timer(); flop = 0;
	local_fft(my_n, my_n/2, (complex_t *)X, TF);
	t = stop_timer();
	barrier();
	if(MYPROC == 0)
		printf("Phase 1 in %.6fs (%.3fMflops/proc)\n", t, flop/(1E6*t));
	barrier();
#if 0	
	start_timer(); flop = 0;
	local_fft(my_n, my_n/2, (complex_t *)X, TF);
	t = stop_timer();
	barrier();
	if(MYPROC == 0)
		printf("Phase 1 in %.6fs (%.3fMflops/proc)\n", t, flop/(1E6*t));
	barrier();
	return;
#endif	
	/* Permute X from cyclic mapping to block mapping */
	if(!terse && MYPROC == 0) printf("Starting remapping\n");
	start_timer();
	all_remap_cyclic_to_blocked(n, X, (void *spread)T);
	t = stop_timer();
	if(MYPROC == 0)
		printf("Remap   in %.6fs (%.3fMb/s/proc)\n", t,
			(my_n*sizeof(complex_t))/(1E6*t));

	/* T is now blocked mapping.
	 * Do log(PROCS) local FFT steps */
	if(!terse && MYPROC == 0)
		printf("Starting %d %d-point local FFTs\n", my_n/PROCS, PROCS);
	start_timer(); flop = 0;
	local_fft(my_n, PROCS/2, (complex_t *)T, TF);
	t = stop_timer();
	if(MYPROC == 0)
		printf("Phase 2 in %.6fs (%.3fMflops/proc)\n", t, flop/(1E6*t));
}

  /* chris: replaced split_main with Mike's */


void splitc_main(int argc, char * argv[])
{
	int n = 1, i;
	complex_t *spread X;
	char c;	
	extern char *optarg;
	
	/* determine and broadcast FFT size */

	/* parse command line args */
	while ((c = getopt(argc, argv, "n:h")) != -1) {
		switch(c) {
		case 'h':
			on_one{
				printf("FFT options: (defaults)\n");
				printf("  -n mil. of keys (in K) per processor  (1 mil)\n");
			}
			exit(0);

		case 'n':  n = atoi(optarg); break;
		}
	}

	n = (1 << 10) * n;	
		
	on_one
	{	
		printf("Doing %d point FFT on %d processors \n", n, PROCS);
		printf("On each processor I have %d points and I allocate %d Kbytes\n", n/PROCS, (n/PROCS*32)>>10);		
	}
	
	/* all_store_sync(); */
	/* n = all_bcast_i(n); */

	/* allocate and init input array */
	X = (complex_t *spread)all_spread_malloc(n, sizeof(complex_t));
	barrier();
	if (!X) {
		if (MYPROC == 0) fprintf(stderr, "Not enough memory!\n");
		barrier();
		exit(1);
	}
	
	for(i=0; i<n/PROCS; i++)
		((complex_t *)X)[i] = (complex_t){0.0, 0.0};
	barrier();
	if(!terse && MYPROC == 0) printf("Initialization done\n");

	/* do FFT */
	all_fft(n, X);
	barrier();

	if(!terse && MYPROC == 0) printf("Done.\n");
}
