/* 

  Basic Global Operations on Blocked Matrixes

*/

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/control.h>

#include "matrix.h"
#include "all_blk_matrix.h"

void all_zero_mat_blk(int r, int c, int b, double A[r][c]::[b][b])
{
  int i,j,l;
  for_my_2D(i,j,l,r,c) {
    double (*lA)[b] = tolocal(A[i][j]);
    zero_matrix(b,b,lA);
  }
  barrier();
}

/* Blocked matrix multiply. Neat thing is take the unblocked
   version, replace gets by bulk gets and we get the blocked version.
*/
void all_mat_mult_blk(int n, int r, int m, int b,
		      double C[n][m]::[b][b], 
		      double A[n][r]::[b][b], 
		      double B[r][m]::[b][b])
{
  int i,j,k,l,ii,jj,kk;
  double la[b][b], lb[b][b];	/* Local copies of the subblocks */


  for_my_2D(i,j,l,n,m) {
    double (*lc)[b] = tolocal(C[i][j]); 
    kk = 0; /*(i+j) % r; */
    for (k=0;k<r;k++) {
      bulk_get (la, A[i][kk], b*b*sizeof(double));
      bulk_get (lb, B[kk][j], b*b*sizeof(double));
      sync();
      matrix_mult(b,b,b,lc,la,lb);
      kk++; if (kk==r) kk=0;
    }
  }
  barrier();
}

/* Just time a matrix multiply without communication. */
void all_local_blk(int n, int r, int m, int b,
		      double C[n][m]::[b][b], 
		      double A[n][r]::[b][b], 
		      double B[r][m]::[b][b])
{
  int i,j,k,l,ii,jj,kk;
  double la[b][b], lb[b][b];	/* Local copies of the subblocks */

  zero_matrix(b,b,la);		/* Clear to avoid exceptions */
  zero_matrix(b,b,lb);

  for_my_2D(i,j,l,n,m) {
    double (*lc)[b] = tolocal(C[i][j]); 
    for (k=0;k<r;k++) {
      matrix_mult(b,b,b,lc,la,lb);
    }
  }
  barrier();
}

/* Just time how it takes to xfer blocks around. */
void all_xfer_blk(int n, int r, int m, int b,
		      double C[n][m]::[b][b], 
		      double A[n][r]::[b][b], 
		      double B[r][m]::[b][b])
{
  int i,j,k,l,ii,jj,kk;
  double la[b][b], lb[b][b];	/* Local copies of the subblocks */


  for_my_2D(i,j,l,n,m) {
    double (*lc)[b] = tolocal(C[i][j]); 
    kk = MYPROC % r;
    for (k=0;k<r;k++) {
      bulk_get (la, A[i][kk], b*b*sizeof(double));
      bulk_get (lb, B[kk][j], b*b*sizeof(double));
      sync();
      kk++;
      if (kk==r) kk=0;
    }
  }
  barrier();
}




