/*									tab:4
 *
 * mm.sc - global matrix multiply based on blocked inner product.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			David E. Culler
 * Version:			31
 * Creation Date:	Wed Jan 13 23:08:30 1993
 * Filename:		mm.sc
 * History:
 *	ssl	2	Mon Oct 25 23:54:07 1993
 *		Fixed scanf. stdin opens in independent mode.
 *	LTL	1	Mon Aug 16 17:43:07 1993
 *		Added code to use scanf for CMMD
 */
static char _version_string_[] = "\nVersion:2:mm.sc\0\n";

#include <stdio.h>
#include <math.h>
#include <split-c/split-c.h>
#include <split-c/com.h>
#include <split-c/control.h>

#include "all_matrix.h"
#include "all_blk_matrix.h"

void all_test_mm (int n, int r, int m, int b)
{
  int i,j,k,l;
  int ii,jj,kk;
  double k2s = 0.0;
  double secs;

  if (b == 0) {
    /* Test Global Matrix Multiplication */
    double (*spread A)[r]::;	/* A[n][r]:: */
    double (*spread B)[m]::;	/* B[r][m]:: */
    double (*spread C)[m]::;	/* C[n][m]:: */

    A = all_spread_malloc(n*r,sizeof(double));
    B = all_spread_malloc(r*m,sizeof(double));
    C = all_spread_malloc(n*m,sizeof(double));
	if (A == 0 || B == 0 || C == 0)
		printf("%d: error: 0 returned by spread malloc\n", MYPROC);

    /* Initialize test data */
    for_my_2D(i,j,l,n,r) A[i][j] = i*j;
    for_my_2D(i,j,l,r,m) B[i][j] = i*j;
    barrier();

    /* Global Matrix Multiply */
	secs = get_seconds();
    all_zero_matrix(n,m,C);
    all_matrix_mult(n,r,m,C,A,B);
	secs = get_seconds() - secs;

    on_one {
      printf("Global Multiply: %6.2f seconds", secs);
      if (secs > 0.001) 
		printf("  %6.2f MFLOPS\n", 2.0*n*m*r/(1000000.0*secs));
      else
		printf("\n");
    }

    /* Check Results */
    /*for (k=0;k<r;k++) k2s = k2s + k*k;
    for_my_2D(i,j,l,n,m) {
      if (fabs((C[i][j] - i*j*k2s)/(i*j*k2s)) > 0.1)
	  {
		  printf("C[%d][%d]=%f, should be %f\n", i,j,C[i][j],i*j*k2s);
		  exit(-1);
	  }
    }*/
    barrier();
  }
  else {

    /* Test Global Blocked Matrix Multiplication */
    double (*spread Ab)[r]::[b][b]; /* Ab[n][r]::[b][b] */
    double (*spread Bb)[m]::[b][b]; /* Bb[r][m]::[b][b] */
    double (*spread Cb)[m]::[b][b]; /* Cb[n][m]::[b][b] */

    Ab = all_spread_malloc(n*r,b*b*sizeof(double));
    Bb = all_spread_malloc(r*m,b*b*sizeof(double));
    Cb = all_spread_malloc(n*m,b*b*sizeof(double));
	if (Ab == 0 || Bb == 0 || Cb == 0)
		printf("%d: error: 0 returned by spread malloc\n", MYPROC);

    /* Initialize test data */
    for_my_2D(i,j,l,n,r) 
      for (ii=0;ii<b;ii++)
		for (jj=0;jj<b;jj++)
		  Ab[i][j][ii][jj] = (i*b+ii)*(j*b+jj);

    for_my_2D(i,j,l,r,m) 
      for (ii=0;ii<b;ii++)
		for (jj=0;jj<b;jj++)
		  Bb[i][j][ii][jj] = (i*b+ii)*(j*b+jj);
    barrier();

	secs = get_seconds();
    all_zero_mat_blk(n,m,b,Cb);
    all_mat_mult_blk(n,r,m,b,Cb,Ab,Bb);
	secs = get_seconds() - secs;

    on_one {
      printf("Global Blocked Multiply: %6.4f seconds", secs);  

      if (secs > 0.001) 
		printf(" %6.2f MFLOPS (%6.2f perPE)\n",
			   2.0*n*m*r*b*b*b/(1000000.0*secs),
			   2.0*n*m*r*b*b*b/(1000000.0*secs*PROCS));
      else
		printf("\n");
    }

    /* Check Results */
    /*k2s = 0.0;
    for (k=0;k<r*b;k++) k2s = k2s + k*k;
    for_my_2D(i,j,l,n,m) {
      for (ii=0;ii<b;ii++)
		for (jj=0;jj<b;jj++) {
		  if ((fabs((Cb[i][j][ii][jj] - (i*b+ii)*(j*b+jj)*k2s)/
					((i*b+ii)*(j*b+jj)*k2s))) > 0.1) {
			char s[1000];
			printf("C[%d][%d]::[%d][%d]=%f, should be %f\n", 
				   i,j,ii,jj,Cb[i][j][ii][jj],(i*b+ii)*(j*b+jj)*k2s, MYPROC);
			exit(-1);
		  }
		}
    }*/
    barrier();

	secs = get_seconds();
    all_zero_mat_blk(n,m,b,Cb);
    all_local_blk(n,r,m,b,Cb,Ab,Bb);
	secs = get_seconds() - secs;


    on_one {
      printf("Local  Blocked Multiply: %6.4f seconds", secs);
      if (secs > 0.001) 
		printf(" %6.2f MFLOPS (%6.2f per PE)\n",
			   2.0*n*m*r*b*b*b/(1000000.0*secs),
			   2.0*n*m*r*b*b*b/(1000000.0*secs*PROCS));
      else
		printf("\n");
    }

	barrier();

	secs = get_seconds();
    all_xfer_blk(n,r,m,b,Cb,Ab,Bb);
	secs = get_seconds() - secs;

    on_one {
      fprintf(stdout, "Transfer blocks:         %6.4f seconds", secs);
      if (secs > 0.001) 
		fprintf(stdout, " %6.2f MBPS (%6.2f per PE)\n",
			   16*n*m*r*b*b/(1000000.0*secs),
			   16*n*m*r*b*b/(1000000.0*secs*PROCS));
      else
		printf("\n");

    }

	barrier();

  }
}
    
splitc_main(int argc, char **argv)
{
  int n,m,r,b;
  int N,M,R,B;
  char c;
  extern char *optarg;
  int i;

#ifdef CM5  
  while ((c = getopt(argc, argv, "s:")) != -1)
  {
	  switch(c)
	  {
	  case 's':
		  all_set_PROCS(atoi(optarg));
		  break;
	  default:
		  printf("\nUsage:\nmm [-s procs]\n");
	  }
  }
#endif  




  on_proc (0) {
    printf("Matrix Multiply: NxR X RxM => NxM\n");
    printf("Enter N R M B:");
    fflush(stdout);
  }
  on_one
	  scanf("%d", &n);
  n = all_bcast_i(n);
  on_one
	  scanf("%d", &r);
  r = all_bcast_i(r);
  on_one
	  scanf("%d", &m);
  m = all_bcast_i(m);
  on_one
	  scanf("%d", &b);
  b = all_bcast_i(b);


  all_test_mm(n,r,m,b);
  barrier();
}
