/*									tab:4
 *
 * mm2x2.c
 *
 * 
 * "Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Richard P. Martin
 * Version:			1
 * Creation Date:	Tue Dec  7 00:02:07 PST 1993
 * Filename:		mm2x2.c
 * History:
 */

/*  mm2x2(C[i][j],A[i][j],B[i][j],r*8,m*8,ksz)
 *
 *  C[i..i+1][j..j+1] := C[i..i+1][j..j+1] 
 *		     + A[i..i+1][k..k+ksz] * B[k..k+ksz][j..j+1]
 *
 */


void mm2x2_inner_k(double *C, double *A,double *B,int r8,int m8, int ksz)
{
  register double c00,c01,c10,c11;
  int i,r,m,rows,col;
  register double *A0,*A1;  /* rows of A*/
  register double *B0,*B1;   /* columns of B */

  r=r8>>3;
  m=m8>>3;

  c00= *C;
  c01= *(C+1);
  c10= *(C+m);
  c11= *(C+m+1);

  A0=A;    /* walk down first row of A */
  A1=A+r;  /* walks down second row of A */

  B0=B;
  B1=B+1;
  
  for (i=0;i<ksz;i++,A0++,A1++,B0+=m,B1+=m){
    c00 += (*A0) * (*B0);
    c01 += (*A0) * (*B1);
    c10 += (*A1) * (*B0);
    c11 += (*A1) * (*B1);
  }

  *C =c00;
  *(C+1) = c01;
  *(C+m) = c10;
  *(C+m+1) = c11;

}
