/*                                  
 * radix.sc - Split-C radix sort program
 */
#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/control.h>
#include <split-c/com.h>
#include "sort.h"

/* 
 * Two key lists are used--on each pass the keys are moved from one to
 * the other, sorting a specific digit (least significant first). 
 */
#define NP_MAX (1024*1024)
unsigned g_keys0[PROCS]::[NP_MAX];
unsigned g_keys1[PROCS]::[NP_MAX]; 
unsigned *keys;
unsigned *keys_aux;

/**
 ** Shoul I do presorting for better splitters
 **/
int DO_PRESORT = 0;

/*
 * DIGITS_PER_KEY * BITS_PER_DIGIT == number of bits in key
 */
#define DIGITS_PER_KEY 2      /* digits in a key (round up) */
#define BITS_PER_DIGIT 16           

/* 
 * Number of buckets == 2^(BITS_PER_DIGIT) 
 */
#define RADIX          65536        
#define RADIX_MASK     (RADIX-1)

/* 
 * A list of buckets provides a count of how many keys on a given
 * processor have a certain value of the digit being sorted. 
 */
unsigned g_buckets[PROCS]::[RADIX];
unsigned *buckets;
/**
 **
 **/
unsigned my_num_keys;



/* 
 * A list of global offsets into the key list allows each processor to
 * write its keys into their new locations. 
 */
int g_scan_buffer[PROCS]::[RADIX];
int *scan_buffer;
int g_scan_result[PROCS]::[RADIX];
int *scan_result;

/**
 ** 
 **/
int send_buffer[RADIX];

/**
 ** buffers, counts, starting positions
 ** for transfer from processor to processor and
 **/
int *my_transfer_buffs[PROCSMAX];
int my_transfer_count[PROCS]::[PROCSMAX];
int *you_transfer_here[PROCS]::[PROCSMAX];

#define MAX_PHASES 40
double phase_time[MAX_PHASES];
double time_finish, time_start;

/**
 ** Names of phases. Number of phases.
 **/
char * phase_name[MAX_PHASES];
int num_phases = 0;

/**
 **
 **/
#if 0
#define NEW_PHASE(name) {phase_time[num_phases] = get_seconds(); phase_name[num_phases++] = name;}
#else
#define NEW_PHASE(name)
#endif


/* 
 * Timing data 
 */
double total_fill, total_hist,total_scan, total_coallesce;
double pass_time[DIGITS_PER_KEY][5];

double last_time, cpu_time, net_time;
#define START_TIME()	{last_time = get_seconds();}
#define SUM_TIME(sum)	{double t=get_seconds(); sum+=t-last_time; last_time=t;}

/*  
 * Procedure: init_buckets
 * Parameters: none
 *
 * Buckets are initialized to 0.
 */
void inline init_buckets()
{
  int i;
    
  for (i = 0; i < RADIX; i++)
    buckets[i] = 0;
}

/*
 * Procedure: histogram
 * Parameters: 
 *   np (input): number of keys per processor
 *   pass (input): the current pass, beginning with 0.
 *   keys (input): local list of keys
 *   
 * Bucket[i] is set to the number of occurrences of the
 * digit i in the list of keys.
 */
void inline histogram(int np, int pass, int *keys) 
{
  int i, shift = pass*BITS_PER_DIGIT;

  for (i = 0; i < np; i++)
    buckets[(keys[i] >> shift) & RADIX_MASK]++;
}

/**
 ** Procedure: compute_transfer_params
 ** Parameters:
 **   np (input): number of keys
 **   a (input):  array of my keys
 **/

void compute_transfer_params(int np, int log_keys, int pass,
				unsigned *in, unsigned *out)
{
  int *l_my_t_count = tolocal(my_transfer_count);
  int i, j, b, tmp, offset;
  int offs[PROCSMAX];
  int shift = pass*BITS_PER_DIGIT;
  int key_bucket, new_proc;
  unsigned aux_buckets[RADIX];


  NEW_PHASE("  BEGIN compute transfer Params");
/****
  for(i = 0; i < np; i++)
     out[i] = in[i];

  for (i = 0; i < RADIX; i++)
   aux_buckets[i] = buckets[i];
****/

  memcpy(out, in, np*sizeof(unsigned));
  memcpy(aux_buckets, buckets, RADIX*sizeof(int));

  NEW_PHASE("    Copy in --> out; buckets --> aux_buckets");

  for(i = 0; i < PROCS; i++)
    l_my_t_count[i] = 0;

   for(i = np - 1; i >= 0 ; i--){
      key_bucket = ((tmp = out[i]) >> shift) & RADIX_MASK; 
      (aux_buckets[key_bucket])--; 
      new_proc = aux_buckets[key_bucket] >> log_keys;
      (l_my_t_count[new_proc])++;
   }

  for(i = 0; i < RADIX; i++)
    if (aux_buckets[i] < 0){
      fprintf(stderr, "%d: Error in aux_buckets (a_b[%d] = %d)!\n", MYPROC,
					i, aux_buckets[i] );
      exit(1);
    } 

  NEW_PHASE("    Compute counts");

  offset = 0;
  for(i = 0; i < PROCS; i++){
    offs[i] = offset;
    offset += l_my_t_count[i];
  }

  for(i = 0; i < PROCS; i++)
    my_transfer_buffs[i] = &(in[offs[i]]);

  for(i = 0; i < PROCS; i++)
      l_my_t_count[i] = 0;

  for(i = np-1; i >= 0 ; i--){
    key_bucket = ((tmp = out[i]) >> shift) & RADIX_MASK; 

    buckets[key_bucket]--; 
    if (key_bucket < 0 ) {
      fprintf(stderr, "key_buck = %d on proc\n", key_bucket, MYPROC);
    }
    new_proc = buckets[key_bucket] >> log_keys;

    my_transfer_buffs[new_proc][l_my_t_count[new_proc]] = tmp; 
    (l_my_t_count[new_proc])++;
  }

  NEW_PHASE("    Copy data to buffers");

/***
  for(i = 0; i < PROCS; i++)
  {
     fprintf(stderr, "%d: count[%d] = %d\n", MYPROC, i, l_my_t_count[i]);
  }
***/

}


  
/**
 ** Compute your_transfer_here
 **/
void compute_your_transfer(unsigned *dest_array, int np)
{
  int proc, offset;
  int **l_you_t_here = tolocal(you_transfer_here);

  barrier();
  for(proc = 0, offset = 0; proc < PROCS; proc++){
    l_you_t_here[proc] = &(dest_array[offset]);
    offset += my_transfer_count[proc][MYPROC];
  }
  my_num_keys = offset;
  
  if (my_num_keys != np){
     fprintf(stderr, "%d: mynumkeys = %d\n", MYPROC, my_num_keys);
     exit(1);
  }

#ifdef ALL_OLD_BARRIERS
   barrier();
#endif
}
 



/**
 ** Procedure: store_block
 ** Parameters:
 **   proc_n (input): destination processor
 **   my_transfer_count:
 **   you_transfer_here:
 **/
void store_block(int proc_n)
{
  int block = 4096;

  int *source_lptr = my_transfer_buffs[proc_n];
  int *global dest_gptr =
    toglobal(proc_n, you_transfer_here[proc_n][MYPROC]);
  int count = my_transfer_count[MYPROC][proc_n];
  int i, size, tmp;

  for(i = 0; i < count; i += block){
    size = block; 
    if ((tmp = (count - i)) < size)
	size = tmp;
    if (size > 0)
      bulk_store(dest_gptr, source_lptr, sizeof(int)*size);
    source_lptr += block;
    dest_gptr += block;
  }
}

/**
 ** Procedure: my_all_to_all_store_blocks
 ** Parameters:
 **   my_transfer_count:
 **   you_transfer_here:
 **/
void my_all_to_all_store_blocks()
{
  int i, n_phases, phase, group_size, half_gsize;
  int pard_halfgroup, pard_halfgroup_offs, pard_i, pard_num;
  int my_halfgroup, my_i, my_half;

  for(i = PROCS, n_phases = 0; (i = i >> 1) != 0; n_phases++)
    ;

  for(phase = 0, group_size = PROCS; phase < n_phases;
      phase++, group_size = group_size >> 1)
  {

    half_gsize = group_size >> 1;
    my_halfgroup = (MYPROC / half_gsize);
    my_i = MYPROC % half_gsize;
    my_half = my_halfgroup % 2;
    pard_halfgroup =  my_half?(my_halfgroup - 1):(my_halfgroup + 1);
    pard_halfgroup_offs = half_gsize*pard_halfgroup;
    for(i = 0; i < half_gsize; i++)
    {
	pard_i = (my_i + i) % half_gsize;
	pard_num = pard_halfgroup_offs + pard_i;
	store_block(pard_num);
        all_store_sync();
    }

  }

  /**
   ** Transfer my own elements. Doing it the lazy way.
   ** Optimize later.
   **/
  store_block(MYPROC);
  all_store_sync();

}



/* 
 * Procedure: all_coalesce 
 * Parameters:
 *   np (input): keys per processor
 *   log_keys (input): log(np)
 *   pass (input): the pass number, beginning with 0
 *   keys (input): the local list of keys that are ranked in
 *                 the buckets array
 *   g_keys (output): the keys after they have been moved to their
 *                    position across processors
 */
void all_coalesce_old(int np, int log_keys, int pass, unsigned *keys, 
            unsigned (* spread g_keys)[NP_MAX])
{
  int i, shift = pass*BITS_PER_DIGIT;
  int key_bucket, new_row, new_column;
  unsigned *global g_ptr;
  int np_mask = np-1;
  int k;

  for (i = np-1; i >= 0; i--) {
    key_bucket = ((k = keys[i]) >> shift) & RADIX_MASK; 
    buckets[key_bucket]--; 
    new_row = buckets[key_bucket] >> log_keys;
    new_column = buckets[key_bucket] & np_mask;
    g_ptr = g_keys[new_row];
    g_ptr += new_column; 
    *g_ptr :- k; 
  }

  all_store_sync();
}


/**
 **
 **/

void new_all_coalesce(int np, int log_keys, int pass, unsigned *in,
			unsigned *out)
{
   int i;
   unsigned tmp1, tmp2;

  NEW_PHASE("BEGIN Coallesce");

  compute_transfer_params(np, log_keys, pass, in, out);
  SUM_TIME(cpu_time);

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

  NEW_PHASE("  END Compute transfer params");

  compute_your_transfer(out, np);

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

  SUM_TIME(net_time);


  NEW_PHASE("  Compute you_transfer");

  my_all_to_all_store_blocks();

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

  SUM_TIME(net_time);

  NEW_PHASE("  All-to-all broadcast");

  if (pass == 0)
  {
     for(i = 0; i < np; i++){
	tmp1 = out[i] << BITS_PER_DIGIT;
	tmp2 = out[i] >> BITS_PER_DIGIT;
	in[i] = tmp1 | tmp2;
      } 
/**
     on_one {
       printf("\n\nThis is in (first 20 elements on proc 1) after swap\n\n");
       for (i=0; i< 20; i++)
         printf("%d ", in[i]);
       printf("\n\n");
     }
**/

     local_radix_sort(np, in, out);

/**
     on_one {
       printf("\n\nThis is the sorted out (first 20 elements on proc 1) before swap\n\n");
       for (i=0; i< 20; i++)
         printf("%d ", out[i]);
       printf("\n\n");
     }
**/
     for(i = 0; i < np; i++){
	tmp1 = out[i] << BITS_PER_DIGIT;
	tmp2 = out[i] >> BITS_PER_DIGIT;
	in[i] = tmp1 | tmp2;
     }

/**     on_one {
       printf("\n\nThis is out (first 20 elements on proc 1) after swap\n\n");
       for (i=0; i< 20; i++)
         printf("%d ", in[i]);
     }
**/

  }
  else if (pass == 1)
    local_radix_sort(np, out, in);
  else
     fprintf(stderr, "%d: Too many passes: %d", MYPROC, pass);

  for(i = 0; i < np; i++)
    out[i] = in[i];

  NEW_PHASE("  Local Radix Sort and swaps");

#ifdef SOME_NEW_BARRIERS 
  barrier();
#endif 

  SUM_TIME(cpu_time);

}



/**
 ** Procedure: all_scan_buckets
 ** Parameters: none
 ** 
 ** On entry, buckets contains the local histogram on each processor. 
 ** On exit, buckets[p][i] contains:
 **    sum buckets[0..PROCS-1][0..(i-1)] +
 **    sum buckets[0..p][i]
 **/ 
void all_scan_buckets()
{
  int i,b,proc;
  int sum = 0;
  int *global neighbor;
  int j, k;

  /**
   ** block size for bulk transfers must be a power of 2
   **/
   int block = 512; 

/****
  fprintf(stderr, "     %2d: all_scan_buckets begin\n", MYPROC);
****/

  /* 
   * Use -1 to designate that a position is empty; works
   * only because keys are only 31 bits.
   */
  for(i = 0; i < RADIX ; i++) {   
    scan_buffer[i] =
    scan_result[i] = -1;
  }
  neighbor = toglobal((MYPROC + 1) % PROCS, scan_buffer);
  barrier(); 

  /*
   * Processor 0 sends each of its RADIX buckets to successive positions
   * on the neighboring processor.  It waits an extra o + tadd between
   * sends so that it doesn't flood the destination processor faster
   * than it can retrieve messages from the network and forward them.
   */
  if (MYPROC == 0) {
    for (b = 0; b < RADIX; b+= block) {
 
      bulk_store(neighbor, &buckets[b], block*sizeof(int));
      neighbor += block;

      /* 
       * Each iteration through the loop takes about ??  usec
       */
       for (j = (block >> 2); j; j--) 
         k += j;
       
    }
  }
  /*
   * The other processors wait until the value has arrived (i.e., it
   * is not -1); it then adds the received value to its local bucket
   * value and, except for the last processor, then forwards the
   * result to its neigbor. The last processor stores the result in
   * the scan_result array.
   */
  else {
    if (MYPROC != (PROCS-1)) {
      for (b = 0; b < RADIX ; b+= block) {
/*****
	store_sync(block*sizeof(int));
*****/
	for(j = 0; j < block; j++)
	{
          /* spin waiting for store to complete */
          while (scan_buffer[b + j] == -1)
	    splitc_poll();
	  send_buffer[b + j] = scan_buffer[b + j] + buckets[b + j];
	}
	bulk_store(neighbor, &send_buffer[b], block*sizeof(int));
        neighbor += block; 
      }
    }
    else {
      for (b = 0; b < RADIX ; b++) {
        /* spin waiting for store to complete */
        while (scan_buffer[b] == -1) splitc_poll();           
        scan_result[b] = scan_buffer[b] + buckets[b];
      }
    }
  }

  barrier();

  NEW_PHASE("  Scan Middle");

  /*
   * The sum of each row of buckets is now broadcast
   * to all of the processors in the scan_result array
   */
  neighbor = toglobal((MYPROC+1) % PROCS, scan_result);

  /*
   * The last processor stores each of the buckets to 
   * processor 0; it waits after each send to avoid
   * flooding processor 0.
   */
  if (MYPROC == (PROCS-1)) {
    for (b = 0; b < RADIX; b+= block) {
      bulk_store(neighbor, &(scan_result[b]), block*sizeof(int));
      neighbor += block;
      for (j = 24; j; j--) 
        k += j;
    } 
  }
  else {
    if (MYPROC < (PROCS-2)) {
      for (b = 0; b < RADIX; b+= block){
	for(j = 0; j < block; j++)
          while (scan_result[b + j] == -1) splitc_poll();
        bulk_store(neighbor, &(scan_result[b]), block*sizeof(int));
        neighbor += block;
      }
    }
    /*
     * The last processor in the pipeline does not
     * forward on the value
     */
    else {
      for (b = 0; b < RADIX; b++) {
        while (scan_result[b] == -1) splitc_poll();
      }
    }
  }

  /*
   * The local value in the buckets array is the sum of all of
   * the previous buckets and the sum of the buckets of this
   * same value on previous processors
   */
  if (MYPROC == 0) {
    for (b = 0; b < RADIX; b++) {
      buckets[b] = sum + buckets[b];
      sum += scan_result[b];
    }
  }
  else {
    for (b = 0; b < RADIX; b++) {
      buckets[b] = sum + buckets[b] + scan_buffer[b];
      sum += scan_result[b];
    }
  }
/****
  fprintf(stderr, "     %2d: all_scan_buckets end\n", MYPROC);
****/
}


void print_results(int np, int distrib)
{
  int i, j;

  on_one {
    printf ("Sort complete.\n\n");
#if 0
    total_fill = total_scan = total_coallesce = 0;
    for (i = 0; i < DIGITS_PER_KEY; i++) { 
      for (j = 4; j > 0; j--)
        pass_time[i][j] -= pass_time[i][j - 1];
      printf ("Pass %d: ",i);
      printf ("Fill: %.4f s Histogram %.4f s Scan: %.3f s  Coallesce: %.3f s\n",
              pass_time[i][1], pass_time[i][2], pass_time[i][3],
              pass_time[i][4]);
      total_fill += pass_time[i][1];
      total_hist += pass_time[i][2];
      total_scan += pass_time[i][3];
      total_coallesce += pass_time[i][4];
    } 

    /* 
     * Print all on one line  
     */
    i = 0;
    printf ("Procs %d, np: %d, passes: %d, distrib: %d\n",
	PROCS, np, DIGITS_PER_KEY, distrib);
    printf ("Total: Fill: %.3fs,  Histo: %.3fs, Scan: %.3fs, "
	" Coallesce: %.3fs Total %.3fs\n",
	total_fill,total_hist,total_scan,total_coallesce,
	time_finish - time_start);
#endif

    printf("@cmd radix-ucsb -n %d\n", np/1024);
    printf ("@cpu %.3f\n@net %.3f\n@total %.3f\n",
	cpu_time, net_time, cpu_time+net_time);
    
    printf ("\nTotal Time: %.3f\n", time_finish - time_start);
    printf ("MSorts per second: %.3f\n", (np*PROCS)/
            ((time_finish - time_start)*1000000));
    printf ("KSorts per second per processor: %.3f\n",
            np/((time_finish - time_start)*1000));
  }
}

/**
 ** Procedure: print_phase_results
 ** Parameter:
 **   np (input): the number of keys per processor
 **   dist (input): entropy parameter
 **/
void print_phase_results(int np, int distrib)
{
  int i;

  on_one {
    printf("\nProcs: %d. n/p: %d. Distribution: %d\n", PROCS, np, distrib);
    for (i = 0; i < num_phases - 1; i++)
      printf("   Phase %2d-%2d:    %.4lf   (%s)\n", i, i+1,
	 phase_time[i+1]-phase_time[i], phase_name[i + 1]);
    printf("Total_Time: %.3f\n", time_finish - time_start);
  }
}



void splitc_main(int argc, char *argv[])
{
  int pass, i, j;
  int np, distrib = 0;
  int procs, log_keys;
  unsigned (* spread g_keys)[NP_MAX];
  int c;
  extern char *optarg;

  np = NP_MAX;
  procs = PROCS;

  /* parse command line args */
  argv++;
  while(--argc) {
    if (!(strncmp(*argv, "-n", 2)))
      np = 1024*atoi(argv[1]);
    if (!(strncmp(*argv, "-d", 2)))
      distrib = atoi(argv[1]);
#ifdef CM5
    if (!(strncmp(*argv, "-p", 2)))
      all_set_PROCS(atoi(argv[1]));
#endif
    if (!(strncmp(*argv, "-h", 2))) {
        on_one{
          printf("Radix Sort options: (defaults)\n");
          printf("  -n Kkeys \n");
          printf("  -p number of procesors\n");
          printf("  -d key distribution\n");
        }
        exit(0);
    }
    argv++;
  }

  /* Calculate log_keys = log(np) */
  log_keys = 0;
  for (i = 1; i < np; i = i<<1)
    log_keys++;

  /* Print out configuration */
  on_one {
    printf ("\nSorting %d 32-bit keys on %d processors (%d keys/proc).\n", 
            np*PROCS, PROCS, np);
    printf ("Radix: %d (%d bits)\n\n", RADIX, BITS_PER_DIGIT);
   fflush(stdout);
  }

#ifdef SOME_NEW_BARRIERS
	barrier();
#endif

  /* Set up local pointers to global arrays. */
  keys = (unsigned *)g_keys0;
  keys_aux = (unsigned *)g_keys1;
  buckets = (unsigned *)g_buckets;
  scan_buffer = (int *)g_scan_buffer;
  scan_result = (int *)g_scan_result;
  create_keys(np, keys, distrib);
  for(i = 0; i < np; i++)
    keys_aux[i] = 0;
  barrier();

/***  
  on_one {
    printf("\n\nThese are the first 20 elements on proc 1 at the beginning\n\n");
    for (i=0; i< 20; i++)
      printf("%d ", keys[i]);
  }
***/

  /* 
   * Sort the numbers, timing individual components.
   * A pass is made over each of the digits of the key.
   */
  START_TIME(); time_start = last_time;
  NEW_PHASE("Start sorting");
  for (pass = 0; pass < DIGITS_PER_KEY; pass++) {

    /*
     * Initialize the buckets
     */
    NEW_PHASE("***** BEGIN a pass *****");
#if 0
    pass_time[pass][0] = get_seconds();
#endif
    keys = (unsigned *) (pass&1 ? tolocal(g_keys1) : tolocal(g_keys0));
    keys_aux = (unsigned *) (pass&1 ? tolocal(g_keys0) : tolocal(g_keys1));
    init_buckets();
    SUM_TIME(cpu_time);

    /* 
     * Count the number of keys with each value of a digit. 
     */
#if 0
    pass_time[pass][1] = get_seconds();
#endif
    NEW_PHASE("Fill");
    histogram(np, pass, keys);
    SUM_TIME(cpu_time);

    /* 
     * Use the bucket histogram to find global offsets. 
     */
#if 0
    pass_time[pass][2] = get_seconds();
#endif
    NEW_PHASE("Build Histogram");
    all_scan_buckets();
    barrier();
    SUM_TIME(net_time);

    /* 
     * Move the keys to their new positions. 
     */
#if 0
    pass_time[pass][3] =  get_seconds();
#endif
    NEW_PHASE("All scan buckets");
    g_keys = (pass&1) ? (unsigned *spread)g_keys0 : (unsigned *spread)g_keys1;
    new_all_coalesce(np, log_keys, pass, keys, keys_aux);

#ifdef ALL_OLD_BARRIERS
	barrier();
#endif

#if 0
    pass_time[pass][4] = get_seconds(); 
#endif
    NEW_PHASE("END Coallesce");
    NEW_PHASE("***** END of pass *****");
  }

  time_finish = get_seconds();

  keys = (unsigned *)(pass & 1 ? tolocal(g_keys1) : tolocal(g_keys0));

#if 0
  on_one {
    printf("\n\nThese are the first 20 elements on proc 1 sorted\n\n");
    for (i=0; i< 20; i++)
      printf("%d ", keys[i]);
  }
#endif

  all_check_results(np, keys, 0);
  print_results(np, distrib);
#if 0
  print_phase_results(np, distrib);
#endif
}




